/*****************************************************************************
 * Copyright (c) 2020 CEA LIST.
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v2.0
 * which accompanies this distribution, and is available at
 * https://www.eclipse.org/legal/epl-2.0/
 *
 * Contributors:
 *  Matteo MORELLI      matteo.morelli@cea.fr - Bug #566899
 *
 *****************************************************************************/

package org.eclipse.papyrus.robotics.ros2.codegen.utils

import org.eclipse.uml2.uml.Package
import org.eclipse.uml2.uml.Class
import org.eclipse.papyrus.robotics.profile.robotics.components.System
import org.eclipse.papyrus.robotics.profile.robotics.skills.SkillDefinition
import org.eclipse.uml2.uml.util.UMLUtil
import java.util.LinkedHashSet
import org.eclipse.uml2.uml.State
import org.eclipse.papyrus.robotics.profile.robotics.skills.SkillOperationalState
import org.eclipse.papyrus.robotics.profile.robotics.skills.SkillSemantic
import static extension org.eclipse.papyrus.robotics.ros2.codegen.utils.Helpers.*
import static extension org.eclipse.papyrus.robotics.ros2.codegen.utils.MessageUtils.*
import static extension org.eclipse.papyrus.robotics.ros2.codegen.utils.PackageTools.pkgName
import org.eclipse.papyrus.robotics.profile.robotics.skills.SkillParameter
import org.eclipse.uml2.uml.DataType

class SkillUtils {

	/**
	 * Determine the file name of C++ code that realizes the skill definition by means of its default semantics
	 */
	def static realizationPackageName(Package pkg) {
		if (pkg=== null)
			throw new Exception("Model error! Can''t access the skill realization model!")
		// otherwise...
		return pkg.pkgName + "_skillrealizations"
	}

	/**
	 * Return the unique skills that the system needs to execute
	 */
	def static LinkedHashSet<SkillDefinition> getUniqueSkills(System sys) {
		var res = new LinkedHashSet<SkillDefinition>
		if (sys !== null)
			for (task : sys.task)
				res.addAll(task.skills)
		return res;
	}

	/**
	 * Get the skill definition name from the base UML::Operation
	 */
	def static getName(SkillDefinition sdef) {
		if (sdef !== null)
			return sdef.base_Operation.name
		return ""
	}

	/**
	 * Return the type of a skill definition parameter as a UML::DataType
	 */
	def static getType(SkillParameter param) {
		if (param !== null)
			if (param.base_Parameter.type instanceof DataType)
				return param.base_Parameter.type as DataType
		return null
	}

	/**
	 * Return the list of unique skill parameter types for a given skill definition
	 */
	def static LinkedHashSet<DataType> getUniqueSkillParameterTypes(SkillDefinition skill) {
		var res = new LinkedHashSet<DataType>
		if (skill !== null)
			for (p : skill.ins + skill.outs)
				res.add(p.type)
		return res;
	}

	/**
	 * Return the ROS2 header file name from a data type (communication object).
	 */
	def static getROS2TypeIncludePath(DataType tp) {
		if (tp !== null)
			return tp.ROS2qMsgName.escapeCamlCase + ".hpp"
		return ""
	}

	/**
	 * Return the name of a skill definition parameter
	 */
	def static getName(SkillParameter param) {
		if (param !== null) {
			return param.base_Parameter.name
		}
		return null
	}

	/**
	 * Return the first operational state of skill FSM
	 */
	def static getFirstOpState(SkillSemantic sem) {
		if (sem !== null)
			return sem.operational.get(0)
		return null
	}

	/**
	 * Check whether a skill operational state will access underlying SW components
	 * by means of a configuration and coordination interface
	 */
	def static doesConfigAndCoordOfComponents(SkillOperationalState ops) {
		if (ops !== null)
			if (ops.compInterface !== null)
				return true;
		return false;
	}

	/**
	 * Return the qualified name of coordination and configuration interface used by a skill operational state
	 */
	def static getCoordinationIfQn(SkillOperationalState ops) {
		if (ops.doesConfigAndCoordOfComponents) {
			val qName = ops.compInterface.base_Interface.ROS2qMsgName
			return qName.replace("/", "::")
		}
		return ""
	}

	/**
	 * Return the qualified name of coordination and configuration interface used by a skill operational state
	 */
	def static getCoordinationIfIncludePath(SkillOperationalState ops) {
		if (ops.doesConfigAndCoordOfComponents) {
			val qName = ops.compInterface.base_Interface.ROS2qMsgName
			return qName.escapeCamlCase + ".hpp"
		}
		return ""
	}

	/**
	 * Return the qualified name of coordination and configuration interface used by a skill operational state
	 */
	def static getCompInterface(SkillOperationalState ops) {
		if (ops.doesConfigAndCoordOfComponents)
			return ops.compInterface.base_Interface
		return null
	}

	/**
	 * Determine the file name of C++ code that realizes the skill definition by means of its default semantics
	 */
	def static realizationFileName(SkillDefinition sdef) {
		if (sdef=== null)
			throw new Exception("Model error! Found UML::Operation with no stereotype applied!")
		// otherwise...
		return String.join("_", sdef.base_Operation.name.split("(?<=.)(?=\\p{Lu})")).toLowerCase + "_action"
	}

	/**
	 * Determine the file name of C++ code that realizes the skill operational state
	 * General version of the function that takes explicitly the SkillSemantic and one of its operational states (NOT USED BY NOW)
	 */
	def static realizationFileName(SkillDefinition sdef, SkillSemantic sem, SkillOperationalState ops) {
		if (sdef=== null)
			throw new Exception("Model error! " + sdef.base_Operation.name + " has no stereotype applied!")
		if (sem === null)
			throw new Exception("Model error! " + sem.base_StateMachine.name + " has no stereotype applied!")
		if (ops === null)
			throw new Exception("Model error! " + ops.base_State.name + " has no stereotype applied!")
		if (!sem.operational.contains(ops))
			throw new Exception("Model error! " + ops.base_State.name + " is not a state of " + sem.base_StateMachine.name)	
		// otherwise...
		var realizBaseName = String.join("_", sdef.base_Operation.name.split("(?<=.)(?=\\p{Lu})")).toLowerCase
		if (sdef.defaultSemantic !== sem)
			realizBaseName += "_" + sem.base_StateMachine.name.toLowerCase + "_" + ops.base_State.name.toLowerCase
		return realizBaseName + "_action_bt_node"
	}

	/**
	 * API that operates on UML object (i.e., not directly on Stereotypes)
	 * ===================================================================
	 */

	/**
	 * doesConfigAndCoordOfComponents() on UML::State
	 */
	def static doesConfigAndCoordOfComponents(State st) {
		return doesConfigAndCoordOfComponents( UMLUtil.getStereotypeApplication(st, SkillOperationalState) )
		
	}

	/**
	 * getUniqueSkills() on UML::Class
	 */
	def static LinkedHashSet<SkillDefinition> getUniqueSkills(Class sys) {
		return getUniqueSkills( UMLUtil.getStereotypeApplication(sys, System) )
	}

}
